% Left-corner parser for Prolog, with oracle, including predicate for
% arc-eager as well as arc-standard parsing.
%
% Written by: Matthew W. Crocker
% Date: 28 April 2000
% Updated: 6 November 2019

% Operator to be used in phrase structure rules.

:- op(1100, xfx, '--->').

% Main predicate for running the test sentences.
do(S,Tree) :-
		test(S,Sentence),write(Sentence),nl,
    parse(Sentence,Tree).

parse(Sentence,Tree) :- parse(s,Sentence,[],Tree,[s],[]),!.

% Parse a Phrase from S1 to S0 beginning with the current left-corner.
parse(Phrase, S1, S0, T,StIn,StOut) :- 
    connects(S1,Word,S2),
    (Cat ---> [Word]),
    link(Cat,Phrase),
    write(StIn), write(' Stack: '), write(S1), nl, 
    write(Cat),write('-->'),write(Word),nl,
    lc(Cat, Phrase, S2, S0, [Cat, [Word]], T,StIn,StOut).

% Compose bottom-up and top-down nodes.
lc(Phrase, Phrase, S0, S0, T, T,[_|S],S).

% Include for arc-eager: early composition of top-down/bottom up nodes.
lc(SubPhrase,Phrase, S1, S0, T1, [Phrase,T1|[TR]],[_|St],StOut) :- 
  (Phrase ---> SubPhrase, Right),
  write(Phrase),write('-->'),write(SubPhrase),write(", "),write(Right),nl,
  parse(Right, S1, S0, TR,[Right|St],StOut).

% General rule: arc standard
lc(SubPhrase,SuperPhrase, S1, S0, T1, Tree,StIn,StOut) :- 
   (Phrase ---> SubPhrase, Right), 
   link(Phrase,SuperPhrase),
   write(Phrase),write('-->'),write(SubPhrase),write(", "),write(Right),nl,
   parse(Right, S1, S2, TR,[Right|StIn],St),
   lc(Phrase, SuperPhrase, S2, S0, [Phrase,T1|[TR]], Tree,St,StOut).

% Match next Word from the difference list.
connects([Word|S], Word, S).

% Grammar

s ---> np, vp.
s ---> np, vi.
np ---> np, pp.
np ---> det, n.
np ---> np, rc.
rc ---> rpro, sgap.
sgap ---> np, vt.
vp ---> vi. %Cannot be used
vp ---> vt, np.
vp ---> vs, s.
vp ---> vpart, part.
pp ---> p, np.

% Lexicon

det ---> [the].
det ---> [her].
rpro ---> [that].
part ---> [down].

n ---> [athlete].
n ---> [dog].
n ---> [cat].
n ---> [goals].
n ---> [medal].
n ---> [leash].
n ---> [medal].

p ---> [with].

vi ---> [smiled].
vi ---> [barked].
vi ---> [walked].

vt ---> [walked].
vt ---> [awarded].
vt ---> [realised].

vs ---> [realised].

vpart ---> [awarded].

% Oracle

%link(s,s). link(np,np). link(det,det). link(rc,rc). link(sgap,sgap). link(vp,vp).
link(np,s). link(det,np). link(det,s). link(rpro,rc). link(np,sgap). 
link(np,det). link(vt,vp). link(vs,vp). link(det,sgap). link(X,X).
link(vpart,vp). link(p,pp).

% Test sentences

test(easy,[the,athlete,walked,the,dog]).

test(npvp,[the,athlete,walked,the,dog,with,the,leash]).

test(nps-np,[the,athlete,realised,her,goals]).
test(nps-s,[the,athlete,realised,her,goals,were,unattainable]).
test(nps-scomp,[the,athlete,realised,that,her,goals,were,unattainable]).

test(cb-zero,[while,the,athlete,walked,the,dog,barked]).
test(cb-zero,[while,the,athlete,smiled,the,dog,barked]).
test(cb-np,[while,the,athlete,walked,the,cat,the,dog,barked]).

test(rr-mc,[the,athlete,awarded,the,medal]).
test(rr-red,[the,athlete,awarded,the,medal,smiled]).
test(rr-unred,[the,athlete,who,was,awarded,the,medal,smiled]).
